from django.db.models.expressions import ExpressionList, F, Func
from django.db.models.indexes import IndexExpression, split_identifier, names_digest
from django.db.models.sql import Query

__all__ = [
    'Index',
    'MinMax', 'Set', 'NgrambfV1', 'TokenbfV1', 'BloomFilter',
]


class Index:
    suffix = 'idx'

    max_name_length = 30

    def __init__(
        self,
        *expressions,
        fields=(),
        name=None,
        type=None,
        granularity=None,
    ):
        if not isinstance(type, Type):
            raise ValueError('Index type must be a Type instance.')
        if not isinstance(granularity, int) and granularity > 0:
            raise ValueError('Index granularity must be a positive integer.')
        if not isinstance(fields, (list, tuple)):
            raise ValueError('Index.fields must be a list or tuple.')
        if not expressions and not fields:
            raise ValueError(
                'At least one field or expression is required to define an '
                'index.'
            )
        if expressions and fields:
            raise ValueError(
                'Index.fields and expressions are mutually exclusive.',
            )
        if expressions and not name:
            raise ValueError('An index must be named to use expressions.')
        if fields and not all(isinstance(field, str) for field in fields):
            raise ValueError('Index.fields must contain only strings with field names.')
        self.fields = list(fields)
        # A list of 2-tuple with the field name and ordering ('' or 'DESC').
        self.fields_orders = [
            (field_name[1:], 'DESC') if field_name.startswith('-') else (field_name, '')
            for field_name in self.fields
        ]
        self.include = []
        self.condition = None
        self.name = name or ''
        self.type = type
        self.granularity = granularity
        self.expressions = tuple(
            F(expression) if isinstance(expression, str) else expression
            for expression in expressions
        )

    @property
    def contains_expressions(self):
        return bool(self.expressions)

    def index_sql(self, model, schema_editor):
        return self.create_sql(model, schema_editor, inline=True)

    def create_sql(self, model, schema_editor, **kwargs):
        if self.expressions:
            index_expressions = []
            for expression in self.expressions:
                index_expression = IndexExpression(expression)
                index_expression.set_wrapper_classes(schema_editor.connection)
                index_expressions.append(index_expression)
            expressions = ExpressionList(*index_expressions).resolve_expression(
                Query(model, alias_cols=False),
            )
        else:
            expressions = None
        return schema_editor._create_index_sql(
            model, fields=self.fields, name=self.name,
            type=self.type, granularity=self.granularity,
            expressions=expressions, **kwargs,
        )

    def remove_sql(self, model, schema_editor, **kwargs):
        return schema_editor._delete_index_sql(model, self.name, **kwargs)

    def deconstruct(self):
        path = '%s.%s' % (self.__class__.__module__, self.__class__.__name__)
        kwargs = {'name': self.name}
        if self.fields:
            kwargs['fields'] = self.fields
        kwargs['type'] = self.type
        kwargs['granularity'] = self.granularity
        return path, self.expressions, kwargs

    def clone(self):
        """Create a copy of this Index."""
        _, args, kwargs = self.deconstruct()
        return self.__class__(*args, **kwargs)

    def set_name_with_model(self, model):
        """
        Generate a unique name for the index.

        The name is divided into 3 parts - table name (12 chars), field name
        (8 chars) and unique hash + suffix (10 chars). Each part is made to
        fit its size by truncating the excess length.
        """
        _, table_name = split_identifier(model._meta.db_table)
        column_names = [model._meta.get_field(field_name).column for field_name in self.fields]
        # The length of the parts of the name is based on the default max
        # length of 30 characters.
        hash_data = [table_name] + column_names + [self.suffix]
        self.name = '%s_%s_%s' % (
            table_name[:11],
            column_names[0][:7],
            '%s_%s' % (names_digest(*hash_data, length=6), self.suffix),
        )
        assert len(self.name) <= self.max_name_length, (
            'Index too long for multiple database support. Is self.suffix '
            'longer than 3 characters?'
        )
        if self.name[0] == '_' or self.name[0].isdigit():
            self.name = 'D%s' % self.name[1:]

    def __repr__(self):
        return '<%s:%s%s%s%s>' % (
            self.__class__.__name__,
            '' if not self.fields else " fields='%s'" % ', '.join(self.fields),
            '' if not self.expressions else " expressions='%s'" % ', '.join([
                str(expression) for expression in self.expressions
            ]),
            ' type=%s' % self.type,
            ' granularity=%s' % self.granularity
        )

    def __eq__(self, other):
        if self.__class__ == other.__class__:
            return self.deconstruct() == other.deconstruct()
        return NotImplemented


class Type(Func):
    pass


class MinMax(Type):
    function = 'minmax'
    template = '%(function)s'
    arity = 0


class Set(Type):
    function = 'set'
    arity = 1


class NgrambfV1(Type):
    function = 'ngrambf_v1'
    arity = 4


class TokenbfV1(Type):
    function = 'tokenbf_v1'
    arity = 3


class BloomFilter(Type):
    function = 'bloom_filter'



from django.db.models.fields import BigAutoField
from django.db.models.sql import compiler

from clickhouse_backend.compat import dj4
from clickhouse_backend.idworker import id_worker


class ClickhouseMixin:
    def as_sql(self, *args, **kwargs):
        sql, params = super().as_sql(*args, **kwargs)
        if self.query.explain_info:
            prefix, suffix = self.connection.ops.explain_query(
                format=self.query.explain_info.format,
                type=self.query.explain_info.type,
                **self.query.explain_info.options,
            )
            sql = "%s %s" % (prefix, sql.lstrip())
            if suffix:
                sql = "%s %s" % (sql, suffix)
        if hasattr(self.query, "setting_info") and self.query.setting_info:
            setting_sql, setting_params = self.connection.ops.settings_sql(
                **self.query.setting_info
            )
            sql = "%s %s" % (sql, setting_sql)
            params = (*params, *setting_params)
        return sql, params


class SQLCompiler(ClickhouseMixin, compiler.SQLCompiler):
    pass


class SQLInsertCompiler(ClickhouseMixin, compiler.SQLInsertCompiler):
    def as_sql(self):
        # We don"t need quote_name_unless_alias() here, since these are all
        # going to be column names (so we can avoid the extra overhead).
        qn = self.connection.ops.quote_name
        opts = self.query.get_meta()
        if dj4:
            insert_statement = self.connection.ops.insert_statement(
                on_conflict=self.query.on_conflict,
            )
        else:
            insert_statement = self.connection.ops.insert_statement(
                ignore_conflicts=self.query.ignore_conflicts
            )

        fields = self.query.fields
        # For compatible with BigAutoField, add value generated by snowflake algorithm if needed.
        absent_of_pk = isinstance(opts.pk, BigAutoField) and opts.pk not in fields
        if absent_of_pk:
            fields = fields + [opts.pk]
            for obj in self.query.objs:
                setattr(obj, opts.pk.attname, id_worker.get_id())

        result = [
            "%s %s(%s)" % (
                insert_statement,
                qn(opts.db_table),
                ", ".join(qn(f.column) for f in fields),
            )
        ]

        value_rows = [
            [self.prepare_value(field, self.pre_save_val(field, obj))
             for field in fields]
            for obj in self.query.objs
        ]

        placeholder_rows, param_rows = self.assemble_as_sql(fields, value_rows)
        result.append(self.connection.ops.bulk_insert_sql(fields, placeholder_rows))
        if hasattr(self.query, "get_settings"):
            settings_string = self.query.get_settings()
            if settings_string:
                result.append(settings_string)
        return [(" ".join(result), param_rows)]

    def execute_sql(self, returning_fields=None):
        as_sql = self.as_sql()
        self.returning_fields = returning_fields
        with self.connection.cursor() as cursor:
            for sql, params in as_sql:
                cursor.execute(sql, params)


class SQLDeleteCompiler(ClickhouseMixin, compiler.SQLDeleteCompiler):
    def _as_sql(self, query):
        """
        When execute DELETE and UPDATE query. Clickhouse does not support
        "table"."column" in WHERE clause.
        """
        table = self.quote_name_unless_alias(query.base_table)
        result = [
            "ALTER TABLE %s DELETE" % table
        ]
        where, params = self.compile(query.where)
        where = where.replace(table + ".", "")
        if where:
            result.append("WHERE %s" % where)
        else:
            result.append("WHERE 1")
        return " ".join(result), tuple(params)


class SQLUpdateCompiler(ClickhouseMixin, compiler.SQLUpdateCompiler):
    def as_sql(self):
        """
        When execute DELETE and UPDATE query. Clickhouse does not support
        "table"."column" in WHERE clause.
        """
        self.pre_sql_setup()
        if not self.query.values:
            return "", ()
        qn = self.quote_name_unless_alias
        values, update_params = [], []
        for field, model, val in self.query.values:
            if hasattr(val, "resolve_expression"):
                val = val.resolve_expression(
                    self.query, allow_joins=False, for_save=True
                )
                if val.contains_aggregate:
                    raise compiler.FieldError(
                        "Aggregate functions are not allowed in this query "
                        "(%s=%r)." % (field.name, val)
                    )
                if val.contains_over_clause:
                    raise compiler.FieldError(
                        "Window expressions are not allowed in this query "
                        "(%s=%r)." % (field.name, val)
                    )
            elif hasattr(val, "prepare_database_save"):
                if field.remote_field:
                    val = field.get_db_prep_save(
                        val.prepare_database_save(field),
                        connection=self.connection,
                    )
                else:
                    raise TypeError(
                        "Tried to update field %s with a model instance, %r. "
                        "Use a value compatible with %s."
                        % (field, val, field.__class__.__name__)
                    )
            else:
                val = field.get_db_prep_save(val, connection=self.connection)

            # Getting the placeholder for the field.
            if hasattr(field, "get_placeholder"):
                placeholder = field.get_placeholder(val, self, self.connection)
            else:
                placeholder = "%s"
            name = field.column
            if hasattr(val, "as_sql"):
                sql, params = self.compile(val)
                values.append("%s = %s" % (qn(name), placeholder % sql))
                update_params.extend(params)
            elif val is not None:
                values.append("%s = %s" % (qn(name), placeholder))
                update_params.append(val)
            else:
                values.append("%s = NULL" % qn(name))

        # Replace "table"."field" to "field", clickhouse does not support that.
        table = qn(self.query.base_table)
        result = [
            "ALTER TABLE %s UPDATE" % table,
            ", ".join(values).replace(table + ".", ""),
        ]
        where, params = self.compile(self.query.where)
        where = where.replace(table + ".", "")

        if where:
            result.append("WHERE %s" % where)
        else:
            result.append("WHERE 1")

        params = (*update_params, *params)
        if hasattr(self.query, "setting_info") and self.query.setting_info:
            setting_sql, setting_params = self.connection.ops.settings_sql(
                **self.query.setting_info
            )
            result.append(setting_sql)
            params = (*params, *setting_params)
        return " ".join(result), params


class SQLAggregateCompiler(ClickhouseMixin, compiler.SQLAggregateCompiler):
    pass



import ipaddress

from django.core import validators
from django.db.models import fields
from django.utils.functional import cached_property

from clickhouse_backend.backend.operations import DatabaseOperations

__all__ = [
    'GenericIPAddressField',
    'PositiveSmallIntegerField', 'PositiveIntegerField', 'PositiveBigIntegerField',
]


class GenericIPAddressField(fields.GenericIPAddressField):
    def db_type(self, connection):
        if self.protocol.lower() == 'ipv4':
            return 'IPv4'
        else:
            return 'IPv6'

    def get_db_prep_value(self, value, connection, prepared=False):
        if not prepared:
            value = self.get_prep_value(value)
        if value is None:
            return value
        try:
            value = ipaddress.ip_address(value)
        except ValueError:
            pass
        else:
            if isinstance(value, ipaddress.IPv4Address) and self.protocol.lower() in ['both', 'ipv6']:
                value = ipaddress.IPv6Address('::ffff:%s' % value)
        return value

    def get_prep_value(self, value):
        if value is None:
            return None
        value = str(value)
        if value and ':' in value:
            try:
                return fields.clean_ipv6_address(value, self.unpack_ipv4)
            except fields.exceptions.ValidationError:
                pass
        return value


class PositiveIntegerFieldMixin:
    """
    Make positive integer have correct limitation corresponding to clickhouse uint type.
    """
    @cached_property
    def validators(self):
        # These validators can't be added at field initialization time since
        # they're based on values retrieved from `connection`.
        validators_ = [*self.default_validators, *self._validators]
        internal_type = self.get_internal_type()
        min_value, max_value = DatabaseOperations.integer_field_ranges[internal_type]
        if min_value is not None and not any(
            (
                isinstance(validator, validators.MinValueValidator) and (
                    validator.limit_value()
                    if callable(validator.limit_value)
                    else validator.limit_value
                ) >= min_value
            ) for validator in validators_
        ):
            validators_.append(validators.MinValueValidator(min_value))
        if max_value is not None and not any(
            (
                isinstance(validator, validators.MaxValueValidator) and (
                    validator.limit_value()
                    if callable(validator.limit_value)
                    else validator.limit_value
                ) <= max_value
            ) for validator in validators_
        ):
            validators_.append(validators.MaxValueValidator(max_value))
        return validators_


class PositiveSmallIntegerField(
    PositiveIntegerFieldMixin,
    fields.PositiveSmallIntegerField
):
    pass


class PositiveIntegerField(
    PositiveIntegerFieldMixin,
    fields.PositiveIntegerField
):
    pass


class PositiveBigIntegerField(
    PositiveIntegerFieldMixin,
    fields.PositiveBigIntegerField
):
    pass



from django.db import models
from django.db.migrations import state
from django.db.models import functions
from django.db.models import options
from django.db.models.manager import BaseManager

from .query import QuerySet
from .sql import Query

options.DEFAULT_NAMES = (
    'verbose_name', 'verbose_name_plural', 'db_table', 'ordering',
    'unique_together', 'permissions', 'get_latest_by', 'order_with_respect_to',
    'app_label', 'db_tablespace', 'abstract', 'managed', 'proxy', 'swappable',
    'auto_created', 'index_together', 'apps', 'default_permissions',
    'select_on_save', 'default_related_name', 'required_db_features',
    'required_db_vendor', 'base_manager_name', 'default_manager_name',
    'indexes', 'constraints',
    # Clickhouse features
    'engine',
)
# Also monkey patch state.DEFAULT_NAMES, this makes new option names contained in migrations.
state.DEFAULT_NAMES = options.DEFAULT_NAMES


def as_clickhouse(self, compiler, connection, **extra_context):
    return functions.Random.as_sql(
        self, compiler, connection, function="rand64", **extra_context
    )


functions.Random.as_clickhouse = as_clickhouse


class ClickhouseManager(BaseManager.from_queryset(QuerySet)):
    def get_queryset(self):
        """
        User defined Query and QuerySet class that support clickhouse particular query.
        """
        return self._queryset_class(
            model=self.model,
            query=Query(self.model),
            using=self._db,
            hints=self._hints
        )


class ClickhouseModel(models.Model):
    objects = ClickhouseManager()

    class Meta:
        abstract = True

    def _do_update(self, base_qs, using, pk_val, values, update_fields, forced_update):
        """
         Try to update the model. Return True if the model was updated (if an
         update query was done and a matching row was found in the DB).
         """
        filtered = base_qs.filter(pk=pk_val)
        if not values:
            # We can end up here when saving a model in inheritance chain where
            # update_fields doesn't target any field in current model. In that
            # case we just say the update succeeded. Another case ending up here
            # is a model with just PK - in that case check that the PK still
            # exists.
            return update_fields is not None or filtered.exists()
        return (
                filtered.exists()
                and
                # It may happen that the object is deleted from the DB right after
                # this check, causing the subsequent UPDATE to return zero matching
                # rows. The same result can occur in some rare cases when the
                # database returns zero despite the UPDATE being executed
                # successfully (a row is matched and updated). In order to
                # distinguish these two cases, the object's existence in the
                # database is again checked for if the UPDATE query returns 0.
                (filtered._update(values) > 0 or filtered.exists())
        )



import ipaddress
from datetime import datetime, timezone
from typing import Sequence, Dict, Union

from clickhouse_driver.util import escape

from . import types

Params = Union[Sequence, Dict]


def escape_datetime(item: datetime, context):
    """Escape datetime to valid clickhouse_backend DateTime literal.

    All datetime are cast to DateTime64 with 6 precision.
    https://clickhouse.com/docs/en/sql-reference/data-types/datetime64
    """
    if item.tzinfo is not None:
        item = item.astimezone(timezone.utc)
    time_string = item.strftime('%Y-%m-%d %H:%M:%S.%f')

    if item.tzinfo is not None:
        return "toDateTime64('%s', 6, 'UTC')" % time_string
    return "toDateTime64('%s', 6)" % time_string


def escape_param(item, context, **kwargs):
    if isinstance(item, ipaddress.IPv4Address):
        return "toIPv4('%s')" % item.compressed
    elif isinstance(item, ipaddress.IPv6Address):
        return "toIPv6('%s')" % item.compressed
    elif isinstance(item, datetime):
        return escape_datetime(item, context)
    # "b'\\x00F  \xfe'" ->   # "\\x00F  \\xfe"
    elif isinstance(item, types.Binary):
        item = str(item)[2:-1]
    return escape.escape_param(item, context)


def escape_params(params: Params, context: Dict) -> Params:
    if isinstance(params, Dict):
        escaped = {
            key: escape_param(value, context)
            for key, value in params.items()
        }
    else:
        escaped = tuple(
            escape_param(value, context)
            for value in params
        )

    return escaped



from django.conf import settings
from django.utils.functional import SimpleLazyObject
from django.utils.module_loading import import_string

from .base import BaseIDWorker
from .snowflake import SnowflakeIDWorker

__all__ = ['BaseIDWorker', 'SnowflakeIDWorker', 'id_worker']

id_worker = SimpleLazyObject(
    lambda: import_string(
        getattr(
            settings,
            'CLICKHOUSE_ID_WORKER',
            'clickhouse_backend.idworker.snowflake.snowflake_worker'
        )
    )
)



from clickhouse_driver import defines
from clickhouse_driver.dbapi import (  # NOQA
    apilevel, threadsafety, paramstyle, __all__
)
from clickhouse_driver.dbapi.errors import (  # NOQA
    Warning, Error, DataError, DatabaseError, ProgrammingError, IntegrityError,
    InterfaceError, InternalError, NotSupportedError, OperationalError
)

from .connection import Connection
from .types import (  # NOQA
    Binary
)


def connect(dsn=None, host=None,
            user=defines.DEFAULT_USER, password=defines.DEFAULT_PASSWORD,
            port=defines.DEFAULT_PORT, database=defines.DEFAULT_DATABASE,
            **kwargs):
    """
    Support dict type params in INSERT query and support connection pool.
    """

    if dsn is None and host is None:
        raise ValueError('host or dsn is required')

    return Connection(dsn=dsn, user=user, password=password, host=host,
                      port=port, database=database, **kwargs)



import re

from django.db.backends.base.introspection import (
    BaseDatabaseIntrospection, FieldInfo, TableInfo,
)
from django.utils.functional import cached_property

constraint_pattern = re.compile(
    r'CONSTRAINT (`)?((?(1)(?:[^\\`]|\\.)+|\S+))(?(1)`|) (CHECK .+?),?\n'
)
index_pattern = re.compile(
    r'INDEX (`)?((?(1)(?:[^\\`]|\\.)+|\S+))(?(1)`|) (.+? TYPE ([a-zA-Z_][0-9a-zA-Z_]*)\(.+?\) GRANULARITY \d+)'
)


class DatabaseIntrospection(BaseDatabaseIntrospection):
    # Maps type codes to Django Field types.
    data_types_reverse = {
        # 'String': 'BinaryField',
        # The best type for String is BinaryField, but sometimes you may need TextField.
        'String': 'TextField',
        'Int64': 'BigIntegerField',
        'Int16': 'SmallIntegerField',
        'Int32': 'IntegerField',
        'UInt64': 'PositiveBigIntegerField',
        'UInt16': 'PositiveSmallIntegerField',
        'UInt32': 'PositiveIntegerField',
        'Float32': 'FloatField',
        'Float64': 'FloatField',
        'IPv4': 'GenericIPAddressField',
        'IPv6': 'GenericIPAddressField',
        'Date': 'DateField',
        'Date32': 'DateField',
        'DateTime': 'DateTimeField',
        'UUID': 'UUIDField',
    }

    ignored_tables = []

    def get_field_type(self, data_type, description):
        if data_type.startswith('FixedString'):  # FixedString(20)
            return 'CharField'
        elif data_type.startswith('DateTime64'):
            return 'DateTimeField'
        elif data_type.startswith('Decimal'):
            return 'DecimalField'
        elif data_type.startswith('Nullable'):  # Nullable(Int16)
            return self.get_field_type(data_type[9:-1], description)
        return super().get_field_type(data_type, description)

    def get_table_list(self, cursor):
        """Return a list of table and view names in the current database."""
        cursor.execute("""
            SELECT table_name,
            CASE table_type WHEN 2 THEN 'v' ELSE 't' END
            FROM INFORMATION_SCHEMA.TABLES
            WHERE table_catalog = currentDatabase()
            AND table_type IN (1, 2)
        """)
        return [TableInfo(*row) for row in cursor.fetchall() if row[0] not in self.ignored_tables]

    def get_table_description(self, cursor, table_name):
        """
        Return a description of the table.
        """
        # Query the INFORMATION_SCHEMA.COLUMNS table.
        cursor.execute("""
            SELECT column_name, data_type, NULL, character_maximum_length,
            coalesce(numeric_precision, datetime_precision),
            numeric_scale, is_nullable, column_default, NULL
            FROM INFORMATION_SCHEMA.COLUMNS
            WHERE table_catalog = currentDatabase() AND table_name = %s
        """, [table_name])
        return [
            FieldInfo(*line)
            for line in cursor.fetchall()
        ]

    def get_constraints(self, cursor, table_name):
        """
        Retrieve any constraints and indexes.
        """
        constraints = {}
        # No way to get structured data, parse from SHOW CREATE TABLE.
        # https://clickhouse.com/docs/en/sql-reference/statements/show#show-create-table
        cursor.execute('SHOW CREATE TABLE "%s"' % table_name)
        table_sql, = cursor.fetchone()
        for backtick, name, definition in constraint_pattern.findall(table_sql):
            constraints[name] = {
                "columns": [],
                "primary_key": False,
                "unique": False,
                "foreign_key": None,
                "check": True,
                "index": False,
                "definition": definition,
                "options": None,
            }

        for backtick, name, definition, type_ in index_pattern.findall(table_sql):
            constraints[name] = {
                "columns": [],
                "orders": [],
                "primary_key": False,
                "unique": False,
                "foreign_key": None,
                "check": False,
                "index": True,
                "type": type_,
                "definition": definition,
                "options": None,
            }
        return constraints

    @cached_property
    def settings(self) -> set:
        """
        Get all available settings.
        """
        with self.connection.cursor() as cursor:
            cursor.execute("SELECT name from system.settings")
            rows = cursor.fetchall()
        return {row[0] for row in rows}



import sys

from clickhouse_driver.errors import ErrorCodes
from django.db.backends.base.creation import BaseDatabaseCreation
from django.db.backends.utils import strip_quotes


class DatabaseCreation(BaseDatabaseCreation):

    def _quote_name(self, name):
        return self.connection.ops.quote_name(name)

    def sql_table_creation_suffix(self):
        test_settings = self.connection.settings_dict["TEST"]
        engine = test_settings.get("ENGINE")
        if engine:
            return "ENGINE = %s" % engine
        return ""

    def _database_exists(self, cursor, database_name):
        cursor.execute(
            "SELECT 1 FROM system.databases WHERE name = %s",
            [strip_quotes(database_name)]
        )
        return cursor.fetchone() is not None

    def create_test_db(self, verbosity=1, autoclobber=False, serialize=True, keepdb=False):
        super().create_test_db(verbosity, autoclobber, serialize, keepdb)
        test_settings = self.connection.settings_dict["TEST"]
        if "fake_transaction" in test_settings:
            self.connection.fake_transaction = test_settings["fake_transaction"]

    def _execute_create_test_db(self, cursor, parameters, keepdb=False):
        try:
            if keepdb and self._database_exists(cursor, parameters["dbname"]):
                # If the database should be kept and it already exists, don't
                # try to create a new one.
                return
            super()._execute_create_test_db(cursor, parameters, keepdb)
        except Exception as e:
            if not e.args or getattr(e.args[0], "code", "") != ErrorCodes.DATABASE_ALREADY_EXISTS:
                # All errors except "database already exists" cancel tests.
                self.log("Got an error creating the test database: %s" % e)
                sys.exit(2)
            elif not keepdb:
                # If the database should be kept, ignore "database already
                # exists".
                raise



from collections import namedtuple

from django.db.models.sql import query
from django.db.models.sql import subqueries

from clickhouse_backend.compat import dj4

ExplainInfo = namedtuple("ExplainInfo", ("format", "type", "options"))


class Query(query.Query):
    def __init__(self, model, where=query.WhereNode, alias_cols=True):
        if dj4:
            super().__init__(model, alias_cols)
        else:
            super().__init__(model, where, alias_cols)
        self.setting_info = {}

    def clone(self):
        obj = super().clone()
        obj.settings = self.setting_info.copy()
        return obj

    def explain(self, using, format=None, type=None, **settings):
        q = self.clone()
        q.explain_info = ExplainInfo(format, type, settings)
        compiler = q.get_compiler(using=using)
        return "\n".join(compiler.explain_query())


for query_class in [subqueries.UpdateQuery, subqueries.DeleteQuery]:
    for attr in ['clone', 'explain']:
        setattr(query_class, attr, getattr(Query, attr))



import re
import time

from clickhouse_driver import client

from .escape import escape_params

insert_pattern = re.compile(r'\s*insert\s*into', flags=re.IGNORECASE)


class Client(client.Client):
    def substitute_params(self, query, params, context):
        escaped = escape_params(params, context)
        return query % escaped

    def execute(self, query, params=None, with_column_types=False,
                external_tables=None, query_id=None, settings=None,
                types_check=False, columnar=False):
        """Support dict params for INSERT queries."""
        start_time = time.time()

        with self.disconnect_on_error(query, settings):
            is_insert = insert_pattern.match(query)

            if is_insert:
                rv = self.process_insert_query(
                    query, params, external_tables=external_tables,
                    query_id=query_id, types_check=types_check,
                    columnar=columnar
                )
            else:
                rv = self.process_ordinary_query(
                    query, params=params, with_column_types=with_column_types,
                    external_tables=external_tables,
                    query_id=query_id, types_check=types_check,
                    columnar=columnar
                )
            self.last_query.store_elapsed(time.time() - start_time)
            return rv



from django.db.models.expressions import Value, Func

__all__ = [
    'Engine', 'BaseMergeTree',
    'MergeTree',
    'ReplacingMergeTree', 'GraphiteMergeTree',
    'CollapsingMergeTree', 'VersionedCollapsingMergeTree',
    'SummingMergeTree', 'AggregatingMergeTree',
    'ReplicatedMergeTree',
    'ReplicatedReplacingMergeTree', 'ReplicatedGraphiteMergeTree',
    'ReplicatedCollapsingMergeTree', 'ReplicatedVersionedCollapsingMergeTree',
    'ReplicatedSummingMergeTree', 'ReplicatedAggregatingMergeTree'
]


class Engine(Func):
    pass


class BaseMergeTree(Engine):
    def __init__(self, *expressions, output_field=None, **extra):
        self.order_by = extra.pop('order_by', None)
        assert self.order_by, 'order_by is required by MergeTree family.'
        self.partition_by = extra.pop('partition_by', None)
        self.primary_key = extra.pop('primary_key', None)
        super().__init__(*expressions, output_field=output_field, **extra)


class MergeTree(BaseMergeTree):
    function = 'MergeTree'
    arity = 0


class ReplacingMergeTree(BaseMergeTree):
    function = 'ReplacingMergeTree'


class SummingMergeTree(BaseMergeTree):
    function = 'SummingMergeTree'


class AggregatingMergeTree(BaseMergeTree):
    function = 'AggregatingMergeTree'
    arity = 0


class CollapsingMergeTree(BaseMergeTree):
    function = 'CollapsingMergeTree'
    arity = 1


class VersionedCollapsingMergeTree(BaseMergeTree):
    function = 'CollapsingMergeTree'
    arity = 2


class GraphiteMergeTree(BaseMergeTree):
    function = 'GraphiteMergeTree'
    arity = 1


class ReplicatedMixin:
    def __init__(self, *expressions, **extra):
        if self.arity is not None and len(expressions) != self.arity + 2:
            raise TypeError(
                "'%s' takes exactly %s %s (%s given)" % (
                    self.__class__.__name__,
                    self.arity + 2,
                    "arguments",
                    len(expressions),
                )
            )
        if self.arity is None and len(expressions) < 2:
            raise TypeError(
                "'%s' takes at least %s %s (%s given)" % (
                    self.__class__.__name__,
                    2,
                    "arguments",
                    len(expressions),
                )
            )
        replicated_params = tuple(Value(arg) for arg in self.expressions[:2])
        super().__init__(*expressions[2:], **extra)
        self.expressions = replicated_params + self.expressions


class ReplicatedMergeTree(ReplicatedMixin, MergeTree):
    function = 'ReplicatedMergeTree'


class ReplicatedReplacingMergeTree(ReplicatedMixin, ReplacingMergeTree):
    function = 'ReplicatedReplacingMergeTree'


class ReplicatedSummingMergeTree(ReplicatedMixin, SummingMergeTree):
    function = 'ReplicatedSummingMergeTree'


class ReplicatedAggregatingMergeTree(ReplicatedMixin, AggregatingMergeTree):
    function = 'ReplicatedAggregatingMergeTree'


class ReplicatedCollapsingMergeTree(ReplicatedMixin, CollapsingMergeTree):
    function = 'ReplicatedCollapsingMergeTree'


class ReplicatedVersionedCollapsingMergeTree(ReplicatedMixin, VersionedCollapsingMergeTree):
    function = 'ReplicatedCollapsingMergeTree'


class ReplicatedGraphiteMergeTree(ReplicatedMixin, GraphiteMergeTree):
    function = 'ReplicatedGraphiteMergeTree'



from django.conf import settings
from django.core.exceptions import ImproperlyConfigured
from django.db.backends.base.base import BaseDatabaseWrapper
from django.db.backends.utils import (
    CursorDebugWrapper as BaseCursorDebugWrapper
)
from django.db.utils import DEFAULT_DB_ALIAS
from django.utils.asyncio import async_unsafe

from clickhouse_backend import driver as Database
from .client import DatabaseClient  # NOQA
from .creation import DatabaseCreation  # NOQA
from .features import DatabaseFeatures  # NOQA
from .introspection import DatabaseIntrospection  # NOQA
from .operations import DatabaseOperations  # NOQA
from .schema import DatabaseSchemaEditor  # NOQA
from django.utils.functional import cached_property


class DatabaseWrapper(BaseDatabaseWrapper):
    vendor = 'clickhouse'
    display_name = 'ClickHouse'
    # This dictionary maps Field objects to their associated ClickHouse column
    # types, as strings. Column-type strings can contain format strings; they'll
    # be interpolated against the values of Field.__dict__ before being output.
    # If a column type is set to None, it won't be included in the output.
    data_types = {
        'AutoField': 'Int32',
        'BigAutoField': 'Int64',
        'IPAddressField': 'IPv4',
        'GenericIPAddressField': 'IPv6',
        'BinaryField': 'String',
        'CharField': 'FixedString(%(max_length)s)',
        'DateField': 'Date32',
        'DateTimeField': "DateTime64(6, 'UTC')" if settings.USE_TZ else 'DateTime64(6)',
        'DecimalField': 'Decimal(%(max_digits)s, %(decimal_places)s)',
        'FileField': 'String',
        'FilePathField': 'String',
        'FloatField': 'Float64',
        'IntegerField': 'Int32',
        'BigIntegerField': 'Int64',
        'PositiveBigIntegerField': 'UInt64',
        'PositiveIntegerField': 'UInt32',
        'PositiveSmallIntegerField': 'UInt16',
        'SlugField': 'String',
        'SmallIntegerField': 'Int16',
        'TextField': 'String',
        'UUIDField': 'UUID',
        'BooleanField': 'Int8',
    }
    operators = {
        'exact': '= %s',
        'iexact': '= UPPER(%s)',
        'contains': 'LIKE %s',
        'icontains': 'ILIKE %s',
        'gt': '> %s',
        'gte': '>= %s',
        'lt': '< %s',
        'lte': '<= %s',
        'startswith': 'LIKE %s',
        'endswith': 'LIKE %s',
        'istartswith': 'ILIKE %s',
        'iendswith': 'ILIKE %s',
    }

    # The patterns below are used to generate SQL pattern lookup clauses when
    # the right-hand side of the lookup isn't a raw string (it might be an expression
    # or the result of a bilateral transformation).
    # In those cases, special characters for LIKE operators (e.g. \, *, _) should be
    # escaped on database side.
    #
    # Note: we use str.format() here for readability as '%' is used as a wildcard for
    # the LIKE operator.
    pattern_esc = r"replaceAll(replaceAll(replaceAll({}, '\\', '\\\\'), '%', '\\%'), '_', '\\_')"
    pattern_ops = {
        'contains': "LIKE '%%' || {} || '%%'",
        'icontains': "ILIKE '%%' || {} || '%%'",
        'startswith': "LIKE {} || '%%'",
        'istartswith': "ILIKE {} || '%%'",
        'endswith': "LIKE '%%' || {}",
        'iendswith': "ILIKE '%%' || {}",
    }

    Database = Database
    SchemaEditorClass = DatabaseSchemaEditor
    # Classes instantiated in __init__().
    client_class = DatabaseClient
    creation_class = DatabaseCreation
    features_class = DatabaseFeatures
    introspection_class = DatabaseIntrospection
    ops_class = DatabaseOperations

    def __init__(self, settings_dict, alias=DEFAULT_DB_ALIAS):
        super().__init__(settings_dict, alias)
        # Use fake_transaction control whether using fake transaction.
        # Fake transaction is used in test, prevent other database such as postgresql
        # from flush at the end of each testcase. Only use this feature when you are
        # aware of the effect in TransactionTestCase.
        self.fake_transaction = settings_dict.get('fake_transaction', False)

    @property
    def fake_transaction(self):
        return self._fake_transaction

    @fake_transaction.setter
    def fake_transaction(self, value):
        self._fake_transaction = value
        self.features.fake_transaction = self._fake_transaction
        self.ops.fake_transaction = self._fake_transaction

    def get_connection_params(self):
        settings_dict = self.settings_dict
        if len(settings_dict['NAME'] or '') > self.ops.max_name_length():
            raise ImproperlyConfigured(
                "The database name '%s' (%d characters) is longer than "
                "Clickhouse's limit of %d characters. Supply a shorter NAME "
                "in settings.DATABASES." % (
                    settings_dict['NAME'],
                    len(settings_dict['NAME']),
                    self.ops.max_name_length(),
                )
            )

        conn_params = {
            'host': settings_dict['HOST'] or 'localhost',
            **settings_dict.get('OPTIONS', {}),
        }
        if settings_dict['NAME']:
            conn_params['database'] = settings_dict['NAME']
        if settings_dict['USER']:
            conn_params['user'] = settings_dict['USER']
        if settings_dict['PASSWORD']:
            conn_params['password'] = settings_dict['PASSWORD']
        if settings_dict['PORT']:
            conn_params['port'] = settings_dict['PORT']
        return conn_params

    @async_unsafe
    def get_new_connection(self, conn_params):
        connection = Database.connect(**conn_params)
        return connection

    def init_connection_state(self):
        pass

    @async_unsafe
    def create_cursor(self, name=None):
        return self.connection.cursor()

    def _set_autocommit(self, autocommit):
        pass

    def is_usable(self):
        try:
            # Use a clickhouse_driver cursor directly, bypassing Django's utilities.
            with self.connection.cursor() as cursor:
                cursor.execute('SELECT 1')
        except Database.Error:
            return False
        else:
            return True

    @cached_property
    def ch_version(self):
        with self.temporary_connection() as cursor:
            cursor.execute("SELECT version()")
            row = cursor.fetchone()
        return row[0]

    def get_database_version(self):
        """
        Return a tuple of the database's version.
        E.g. for ch_version "22.9.3.18", return (22, 9, 3, 18).
        """
        return tuple(map(int, self.ch_version.split(".")))



from clickhouse_driver.dbapi import connection
from clickhouse_driver.dbapi import cursor

from .pool import ClickhousePool


class Cursor(cursor.Cursor):
    def close(self):
        """Push client back to connection pool"""
        self._state = self._states.CURSOR_CLOSED
        self._connection.pool.push(self._client)

    @property
    def closed(self):
        return self._state == self._states.CURSOR_CLOSED


class Connection(connection.Connection):
    """Connection class with support for connection pool."""
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.pool = ClickhousePool(
            connections_min=0, connections_max=500,
            dsn=self.dsn, host=self.host, port=self.port,
            user=self.user, password=self.password,
            database=self.database, **self.connection_kwargs,
        )

    def _make_client(self):
        """
        :return: a new Client instance.
        """
        return self.pool.pull()

    def close(self):
        self.pool.cleanup()
        self.is_closed = True

    def cursor(self, cursor_factory=Cursor):
        """Use clickhouse_backend.connection.Cursor which support
        connection pool to create cursor."""
        return super().cursor(cursor_factory)



from pathlib import Path

from setuptools import setup, find_packages


def read(name):
    with open((Path(__file__).parent / name)) as file:
        return file.read()


setup(
    name='django-clickhouse-backend',
    version=read('clickhouse_backend/VERSION').strip(),
    description='Django clickHouse database backend.',
    packages=find_packages(),
    include_package_data=True,
    url='https://github.com/jayvynl/django-clickhouse-backend',
    author='Lin Zhiwen',
    author_email='zhiwenlin1116@gmail.com',
    long_description=read('README.md'),
    long_description_content_type='text/markdown',
    license='MIT',
    install_requires=[
        'django>=3.2',
        'clickhouse-driver==0.2.4',
        'clickhouse-pool==0.5.3',
    ],
)
